package com.sd.business.service.production.impl;

import cn.hutool.core.util.ObjectUtil;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.ruoyi.common.constant.StatusConstant;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.ruoyi.common.utils.SecurityUtils;
import com.ruoyi.common.utils.wrapper.IWrapper;
import com.sd.business.entity.bom.bo.BomSpecBo;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.department.constant.DepartmentConstant;
import com.sd.business.entity.department.po.Department;
import com.sd.business.entity.in.dto.InOutStorageDto;
import com.sd.business.entity.in.emums.InOutTypeEnum;
import com.sd.business.entity.in.emums.OutDetailTypeEnum;
import com.sd.business.entity.in.po.InOutStorageBom;
import com.sd.business.entity.inventory.po.Inventory;
import com.sd.business.entity.order.enums.OrderExceptionTypeEnum;
import com.sd.business.entity.order.enums.OrderStatusEnum;
import com.sd.business.entity.order.po.OrderInfo;
import com.sd.business.entity.order.po.OrderSku;
import com.sd.business.entity.order.po.OrderSkuBom;
import com.sd.business.entity.production.dto.StockPreparationDto;
import com.sd.business.entity.production.po.ProductionTask;
import com.sd.business.entity.production.po.ProductionWorkOrder;
import com.sd.business.entity.production.vo.OutBomVo;
import com.sd.business.entity.production.vo.PackageBomVo;
import com.sd.business.entity.production.vo.StockPreparationVo;
import com.sd.business.entity.production.vo.UncompletedVo;
import com.sd.business.entity.sku.po.Sku;
import com.sd.business.entity.sku.po.SkuSpec;
import com.sd.business.entity.warehouse.constant.WarehouseConstant;
import com.sd.business.mapper.production.StockPreparationMapper;
import com.sd.business.service.bom.BomClassifyService;
import com.sd.business.service.bom.BomService;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.in.InOutStorageService;
import com.sd.business.service.inventory.InventoryService;
import com.sd.business.service.order.OrderService;
import com.sd.business.service.order.OrderSkuBomService;
import com.sd.business.service.order.OrderSkuService;
import com.sd.business.service.production.ProductionTaskService;
import com.sd.business.service.production.ProductionWorkOrderService;
import com.sd.business.service.production.StockPreparationService;
import com.sd.business.service.sku.SkuService;
import com.sd.business.service.sku.SkuSpecService;
import com.sd.framework.util.Assert;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Service
public class StockPreparationServiceImpl implements StockPreparationService {

    @Resource
    private StockPreparationMapper stockPreparationMapper;

    @Autowired
    private OrderService orderService;

    @Autowired
    private OrderSkuService orderSkuService;

    @Autowired
    private OrderSkuBomService orderSkuBomService;

    @Autowired
    private ProductionTaskService productionTaskService;

    @Autowired
    private ProductionWorkOrderService productionWorkOrderService;

    @Autowired
    private SkuService skuService;

    @Autowired
    private SkuSpecService skuSpecService;

    @Autowired
    private InOutStorageService inOutStorageService;

    @Autowired
    private BomSpecService bomSpecService;

    @Autowired
    private BomService bomService;

    @Autowired
    private BomClassifyService bomClassifyService;

    @Autowired
    private InventoryService inventoryService;

    @Override
    public List<UncompletedVo> uncompletedList(StockPreparationDto dto) {

        IWrapper<StockPreparationVo> wrapper = getWrapper(dto);
        List<StockPreparationVo> stockPreparationVoList = stockPreparationMapper.uncompletedList(wrapper);

        return new ArrayList<>(stockPreparationVoList.stream().collect(Collectors.toMap(
                StockPreparationVo::getBomSpecId,
                item -> {
                    UncompletedVo uncompletedVo = new UncompletedVo();
                    uncompletedVo.setBomSpecId(item.getBomSpecId());
                    uncompletedVo.setBomSpecCode(item.getBomSpecCode());
                    uncompletedVo.setBomSpecName(item.getBomSpecName());
                    uncompletedVo.setTotalQuantity(item.getQuantity());

                    UncompletedVo.SkuInfo skuInfo = new UncompletedVo.SkuInfo();
                    skuInfo.setSkuSpecId(item.getSkuSpecId());
                    skuInfo.setSkuSpecCode(item.getSkuSpecCode());
                    skuInfo.setSkuSpecName(item.getSkuSpecName());
                    skuInfo.setQuantity(item.getQuantity());

                    List<UncompletedVo.SkuInfo> skuInfoList = new ArrayList<>();
                    skuInfoList.add(skuInfo);

                    uncompletedVo.setSkuInfoList(skuInfoList);

                    return uncompletedVo;
                },
                (v1, v2) -> {
                    v1.setTotalQuantity(v1.getTotalQuantity().add(v2.getTotalQuantity()));
                    UncompletedVo.SkuInfo skuInfo = v2.getSkuInfoList().get(0);
                    boolean flag = true;
                    for (UncompletedVo.SkuInfo temp : v1.getSkuInfoList()) {
                        if (Objects.equals(temp.getSkuSpecId(), skuInfo.getSkuSpecId())) {
                            temp.setQuantity(temp.getQuantity().add(skuInfo.getQuantity()));
                            flag = false;
                            break;
                        }
                    }

                    if (flag) {
                        v1.getSkuInfoList().add(skuInfo);
                    }

                    return v1;
                }
        )).values());

    }

    @Override
    public Page<StockPreparationVo> completedPage(StockPreparationDto dto) {
        IWrapper<StockPreparationVo> wrapper = IWrapper.getWrapper();
        wrapper.eq("oi", OrderInfo::getCode, dto.getOrderCode());
        wrapper.eq("oi", OrderInfo::getWlnCode, dto.getOrderWlnCode());
        wrapper.eq("bs", BomSpec::getCode, dto.getBomSpecCode());
        wrapper.eq("bs", BomSpec::getName, dto.getBomSpecName());
        wrapper.eq("d", Department::getId, dto.getDepartmentId());
        wrapper.eq("ss", SkuSpec::getCode, dto.getSkuSpecCode());
        wrapper.eq("ss", SkuSpec::getName, dto.getSkuSpecName());
        wrapper.eq("bs", BomSpec::getWidth, dto.getWidth());
        wrapper.eq("os", OrderSku::getStockPreparationStatus, StatusConstant.YES);
        wrapper.orderByDesc("os", OrderSku::getStockPreparationTime);
        wrapper.orderByAsc("os", OrderSku::getId);
        return stockPreparationMapper.completedPage(dto.getPage(), wrapper);
    }

    @Transactional(rollbackFor = Exception.class)
    @Override
    public synchronized void submit(List<Long> orderSkuIdList) {

        // 更爱状态为已备料
        List<OrderSku> orderSkuList = orderSkuService.list(q -> q
                .eq(OrderSku::getStockPreparationStatus, StatusConstant.NO)
                .in(BaseIdPo::getId, orderSkuIdList));

        if (orderSkuList.size() == 0) {
            return;
        }

        // 订单sku出库
        Date stockPreparationTime = new Date();
        orderSkuList.forEach(item -> {
            item.setStockPreparationTime(stockPreparationTime);
            item.setStockPreparationStatus(StatusConstant.YES);
        });
        orderSkuService.updateBatchById(orderSkuList);

        // 订单id列表
        List<Long> orderIdList = orderSkuList.stream().map(OrderSku::getOrderId).distinct().collect(Collectors.toList());

        // 订单出库
        orderDelivery(orderIdList, orderSkuList);

        List<OrderInfo> orderInfoList = new ArrayList<>();
        List<ProductionTask> productionTaskList = new ArrayList<>();
        List<ProductionWorkOrder> productionWorkOrderList = new ArrayList<>();

        // 生产任务、生产工单
        Map<Long, List<OrderSku>> map = orderSkuList.stream().collect(Collectors.groupingBy(OrderSku::getOrderId));
        map.forEach((orderId, itemOrderSkuList) -> {
            // 订单状态改为 生产中
            addOrderInfo(orderId, orderInfoList);
            // 添加 生产任务、工单
            addTaskAndWorkOrder(orderId, itemOrderSkuList, productionTaskList, productionWorkOrderList);
        });

        // 保存数据
        if (orderInfoList.size() > 0) {
            orderService.updateBatchById(orderInfoList);
        }
        if (productionTaskList.size() > 0) {
            productionTaskService.saveBatch(productionTaskList);
        }
        if (productionWorkOrderList.size() > 0) {
            productionWorkOrderService.saveBatch(productionWorkOrderList);
        }

    }

    @Override
    public List<PackageBomVo> getPackageBomList(List<Long> orderSkuIdList) {

        List<OrderSkuBom> list = orderSkuBomService.list(q -> q.in(OrderSkuBom::getOrderSkuId, orderSkuIdList));

        List<PackageBomVo> result = new ArrayList<>(list.stream()
                .map(item -> PackageBomVo.builder().bomSpecId(item.getBomSpecId()).quantity(item.getQuantity()).build())
                .collect(Collectors.toMap(
                        PackageBomVo::getBomSpecId,
                        Function.identity(),
                        (v1, v2) -> {
                            v1.setQuantity(v1.getQuantity().add(v2.getQuantity()));
                            return v1;
                        }
                )).values());

        bomSpecService.attributeAssign(result, PackageBomVo::getBomSpecId, (item, bomSpec) -> {
            item.setBomSpecCode(bomSpec.getCode());
            item.setBomSpecName(bomSpec.getName());
            item.setBomId(bomSpec.getBomId());
        });

        bomService.attributeAssign(result, PackageBomVo::getBomId, (item, bom) -> {
            item.setBomClassifyId(bom.getBomClassifyId());
        });

        bomClassifyService.attributeAssign(result, PackageBomVo::getBomClassifyId, (item, bomClassify) -> {
            item.setBomClassifyName(bomClassify.getName());
        });

        return result;
    }

    @Override
    public List<OutBomVo> outBomList(StockPreparationDto dto) {
        Long outDepartmentId = dto.getOutDepartmentId();
        Assert.notNull(outDepartmentId, "出库事业部id不能为空");

        // 订单sku列表
        IWrapper<StockPreparationVo> wrapper = getWrapper(dto);
        List<StockPreparationVo> stockPreparationVoList = stockPreparationMapper.uncompletedList(wrapper);

        // 订单sku包材列表
        List<Long> orderSkuIdList = stockPreparationVoList.stream().map(StockPreparationVo::getOrderSkuId).collect(Collectors.toList());
        List<OrderSkuBom> orderSkuBomList = orderSkuBomService.listByIds(orderSkuIdList);

        // 合并列表
        List<OutBomVo> outBomVoList = stockPreparationVoList.stream().map(item -> {
            OutBomVo outBomVo = new OutBomVo();
            outBomVo.setBomSpecId(item.getBomSpecId());
            outBomVo.setOutQuantity(item.getQuantity());
            return outBomVo;
        }).collect(Collectors.toList());

        for (OrderSkuBom orderSkuBom : orderSkuBomList) {
            OutBomVo outBomVo = new OutBomVo();
            outBomVo.setBomSpecId(orderSkuBom.getBomSpecId());
            outBomVo.setOutQuantity(orderSkuBom.getQuantity());
            outBomVoList.add(outBomVo);
        }

        // 合并数量
        Map<Long, OutBomVo> map = outBomVoList.stream().collect(Collectors.toMap(
                OutBomVo::getBomSpecId,
                Function.identity(),
                (v1, v2) -> {
                    v1.setOutQuantity(v1.getOutQuantity().add(v2.getOutQuantity()));
                    return v1;
                }
        ));
        outBomVoList = new ArrayList<>(map.values());

        // 区分半成品仓和包材仓库
        Map<Long, BomSpecBo> bomSpecBoMap = skuSpecService.getBomSpecBoByIdList(map.keySet());

        for (OutBomVo outBomVo : outBomVoList) {
            Long bomSpecId = outBomVo.getBomSpecId();
            BomSpecBo bomSpecBo = bomSpecBoMap.get(bomSpecId);
            outBomVo.setBomSpecCode(bomSpecBo.getBomSpecCode());
            outBomVo.setBomSpecName(bomSpecBo.getBomSpecName());
            // 主材
            if (Objects.equals(bomSpecBo.getClassifyParentId(), 1L)) {
                outBomVo.setWarehouseId(WarehouseConstant.SEMI_FINISHED_PRODUCT);
                outBomVo.setWarehouseName("半成品仓");
            }
            // 包材
            else {
                outBomVo.setWarehouseId(WarehouseConstant.PACKAGING_MATERIAL);
                outBomVo.setWarehouseName("包材仓");
            }
        }

        // 主材库存
        Map<Long, BigDecimal> semiFinishedProductInventoryMap = Collections.emptyMap();
        List<Long> semiFinishedProductBomSpecIdList = outBomVoList.stream()
                .filter(item -> WarehouseConstant.SEMI_FINISHED_PRODUCT.equals(item.getWarehouseId()))
                .map(OutBomVo::getBomSpecId)
                .collect(Collectors.toList());
        if (semiFinishedProductBomSpecIdList.size() > 0) {
            semiFinishedProductInventoryMap = inventoryService.mapKV(
                    Inventory::getBomSpecId,
                    Inventory::getQuantity,
                    q -> q.eq(Inventory::getWarehouseId, WarehouseConstant.SEMI_FINISHED_PRODUCT)
                            .eq(Inventory::getDepartmentId, outDepartmentId)
                            .in(Inventory::getBomSpecId, semiFinishedProductBomSpecIdList)
            );
        }

        // 包材库存
        Map<Long, BigDecimal> packagingMaterialInventoryMap = Collections.emptyMap();
        List<Long> packagingMaterialBomSpecIdList = outBomVoList.stream()
                .filter(item -> WarehouseConstant.PACKAGING_MATERIAL.equals(item.getWarehouseId()))
                .map(OutBomVo::getBomSpecId)
                .collect(Collectors.toList());
        if (packagingMaterialBomSpecIdList.size() > 0) {
            packagingMaterialInventoryMap = inventoryService.mapKV(
                    Inventory::getBomSpecId,
                    Inventory::getQuantity,
                    q -> q.eq(Inventory::getWarehouseId, WarehouseConstant.PACKAGING_MATERIAL)
                            .eq(Inventory::getDepartmentId, outDepartmentId)
                            .in(Inventory::getBomSpecId, semiFinishedProductBomSpecIdList)
            );
        }

        // 赋值库存
        for (OutBomVo outBomVo : outBomVoList) {
            if (WarehouseConstant.SEMI_FINISHED_PRODUCT.equals(outBomVo.getWarehouseId())) {
                outBomVo.setInventoryQuantity(semiFinishedProductInventoryMap.getOrDefault(outBomVo.getBomSpecId(), BigDecimal.ZERO));
            } else {
                outBomVo.setInventoryQuantity(packagingMaterialInventoryMap.getOrDefault(outBomVo.getBomSpecId(), BigDecimal.ZERO));
            }
        }

        return outBomVoList.stream()
                .sorted(Comparator.comparing(OutBomVo::getWarehouseId, Comparator.reverseOrder()).thenComparing(OutBomVo::getBomSpecCode))
                .collect(Collectors.toList());

    }

    /**
     * 获取wrapper
     */
    private IWrapper<StockPreparationVo> getWrapper(StockPreparationDto dto) {
        Assert.notNull(dto.getDepartmentId(), "事业部id不能为空");

        IWrapper<StockPreparationVo> wrapper = IWrapper.getWrapper();
        wrapper.eq("oi", OrderInfo::getStatus, OrderStatusEnum.STOCK_PREPARATION.getKey());
        wrapper.eq("oi", OrderInfo::getExceptionType, OrderExceptionTypeEnum.NORMAL.getKey().toString());
        wrapper.eq("os", OrderSku::getStockPreparationStatus, StatusConstant.NO);
        wrapper.eq("oi", OrderInfo::getDepartmentId, dto.getDepartmentId());
        wrapper.eq("bs", BomSpec::getCode, dto.getBomSpecCode());
        wrapper.eq("bs", BomSpec::getName, dto.getBomSpecName());
        wrapper.eq("ss", SkuSpec::getCode, dto.getSkuSpecCode());
        wrapper.eq("ss", SkuSpec::getName, dto.getSkuSpecName());
        return wrapper;
    }

    /**
     * 订单状态改为 生产中
     */
    private void addOrderInfo(Long orderId, List<OrderInfo> orderInfoList) {
        OrderInfo orderInfo = new OrderInfo();
        orderInfo.setId(orderId);
        orderInfo.setStatus(OrderStatusEnum.IN_PRODUCTION.getKey());
        orderInfoList.add(orderInfo);
    }

    /**
     * 添加 生产任务、工单
     */
    private void addTaskAndWorkOrder(Long orderId,
                                     List<OrderSku> orderSkuList,
                                     List<ProductionTask> productionTaskList,
                                     List<ProductionWorkOrder> productionWorkOrderList) {

        OrderInfo orderInfo = orderService.getById(orderId);
        Map<String, ProductionTask> skuSpecIdProductionTaskMap = new HashMap<>();
        int num = 1;

        for (OrderSku orderSku : orderSkuList) {

            Long skuSpecId = orderSku.getSkuSpecId();
            Long bomSpecId = orderSku.getBomSpecId();
            Integer printType = orderSku.getPrintType();
            BigDecimal quantity = orderSku.getQuantity();

            ProductionTask productionTask = skuSpecIdProductionTaskMap.computeIfAbsent(
                    skuSpecId + ":" + bomSpecId + ":" + printType,
                    item -> {
                        SkuSpec skuSpec = skuSpecService.getById(skuSpecId);
                        Sku sku = skuService.getById(skuSpec.getSkuId());

                        ProductionTask tempProductionTask = new ProductionTask();
                        tempProductionTask.setId(IdWorker.getId());
                        tempProductionTask.setOrderId(orderSku.getOrderId());
                        tempProductionTask.setSkuSpecId(skuSpecId);
                        tempProductionTask.setBomSpecId(bomSpecId);
                        tempProductionTask.setCraftProductionLineId(sku.getCraftProductionLineId());
                        tempProductionTask.setMachinedPanel(skuSpec.getMachinedPanel());
                        tempProductionTask.setPrintType(printType);
                        tempProductionTask.setProductionQuantity(BigDecimal.ZERO);
                        tempProductionTask.setCompleteQuantity(BigDecimal.ZERO);

                        productionTaskList.add(tempProductionTask);

                        return tempProductionTask;
                    });

            productionTask.setProductionQuantity(productionTask.getProductionQuantity().add(quantity));

            for (int i = 0; i < quantity.intValue(); i++) {
                ProductionWorkOrder productionWorkOrder = new ProductionWorkOrder();
                productionWorkOrder.setProductionTaskId(productionTask.getId());
                productionWorkOrder.setCode(orderInfo.getCode() + String.format("-%05d", num));
                productionWorkOrder.setStatus(0);
                productionWorkOrderList.add(productionWorkOrder);
                num++;
            }

        }
    }

    /**
     * 订单出库
     */
    private void orderDelivery(List<Long> orderIdList, List<OrderSku> orderSkuList) {

        List<OrderSkuBom> orderSkuBomList = orderSkuBomService.list(q -> q.in(OrderSkuBom::getOrderId, orderIdList));

        // 合并sku主材和包材
        List<InOutStorageBom> list = new ArrayList<>(
                Stream.concat(
                                orderSkuList.stream().map(item -> {
                                    InOutStorageBom inOutStorageBom = new InOutStorageBom();
                                    inOutStorageBom.setBomSpecId(item.getBomSpecId());
                                    inOutStorageBom.setQuantity(item.getQuantity());
                                    return inOutStorageBom;
                                }),
                                orderSkuBomList.stream().map(item -> {
                                    InOutStorageBom inOutStorageBom = new InOutStorageBom();
                                    inOutStorageBom.setBomSpecId(item.getBomSpecId());
                                    inOutStorageBom.setQuantity(item.getQuantity());
                                    return inOutStorageBom;
                                })
                        )
                        .collect(Collectors.toMap(
                                        InOutStorageBom::getBomSpecId,
                                        Function.identity(),
                                        (v1, v2) -> {
                                            v1.setQuantity(v1.getQuantity().add(v2.getQuantity()));
                                            return v1;
                                        }
                                )
                        ).values());

        List<Long> bomSpecIdList = list.stream().map(InOutStorageBom::getBomSpecId).collect(Collectors.toList());
        Map<Long, BomSpecBo> bomSpecBo = skuSpecService.getBomSpecBoByIdList(bomSpecIdList);

        // 主材从半成品仓出库
        InOutStorageDto semiFinishedProduct = new InOutStorageDto();
        semiFinishedProduct.setType(InOutTypeEnum.OUT.getKey());
        semiFinishedProduct.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        semiFinishedProduct.setWarehouseId(WarehouseConstant.SEMI_FINISHED_PRODUCT);
        semiFinishedProduct.setDepartmentId(DepartmentConstant.SD_SPORTS);
        semiFinishedProduct.setApplicant(SecurityUtils.getLoginUser().getUser().getNickName());
        List<InOutStorageBom> semiFinishedProductInOutStorageBomList = list.stream().filter(item -> {
            BomSpecBo tempBomSpecBo = bomSpecBo.get(item.getBomSpecId());
            if (tempBomSpecBo == null) {
                return false;
            }
            return ObjectUtil.equal(tempBomSpecBo.getClassifyParentId(), 1L);
        }).collect(Collectors.toList());
        semiFinishedProduct.setInOutStorageBomList(semiFinishedProductInOutStorageBomList);
        inOutStorageService.add(semiFinishedProduct);

        // 包材从包材仓出库
        InOutStorageDto packagingMaterial = new InOutStorageDto();
        packagingMaterial.setType(InOutTypeEnum.OUT.getKey());
        packagingMaterial.setDetailType(OutDetailTypeEnum.PRODUCTION.getKey());
        packagingMaterial.setWarehouseId(WarehouseConstant.PACKAGING_MATERIAL);
        packagingMaterial.setDepartmentId(DepartmentConstant.SD_SPORTS);
        packagingMaterial.setApplicant(SecurityUtils.getLoginUser().getUser().getNickName());
        List<InOutStorageBom> packagingMaterialInOutStorageBomList = list.stream().filter(item -> {
            BomSpecBo tempBomSpecBo = bomSpecBo.get(item.getBomSpecId());
            if (tempBomSpecBo == null) {
                return true;
            }
            return ObjectUtil.notEqual(tempBomSpecBo.getClassifyParentId(), 1L);
        }).collect(Collectors.toList());
        packagingMaterial.setInOutStorageBomList(packagingMaterialInOutStorageBomList);
        inOutStorageService.add(packagingMaterial);

    }

}
